# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import os

import pylook


# -- Project information -----------------------------------------------------

project = 'pylook'
copyright = '2020, Leeman Geophysical LLC'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx_gallery.gen_gallery',
    'sphinx.ext.napoleon'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
#html_theme = 'alabaster'
try:
    import sphinx_rtd_theme
    html_theme = 'sphinx_rtd_theme'
    html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
except ImportError:
    pass

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_css_files = ['theme_override.css']

# Controlling automatically generating summary tables in the docs
autosummary_generate = True
autosummary_imported_members = True

# Control main class documentation
autoclass_content = 'both'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
verinfo = pylook.__version__
full_version = verinfo.split('+')[0]
version = full_version.rsplit('.', 1)[0]
# The full version, including alpha/beta/rc tags.
release = verinfo

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

sphinx_gallery_conf = {
    'doc_module': ('pylook',),
    'reference_url': {
        'pylook': None,
    },
    'examples_dirs': [os.path.join('..', 'examples')],
    'gallery_dirs': ['examples'],
    'filename_pattern': r'\.py',
    'backreferences_dir': os.path.join('api', 'generated'),
    #'default_thumb_file': os.path.join('_static', 'pylook_150x150_white_bg.png'),
    'abort_on_example_error': True
}

# Tweak how docs are formatted
napoleon_use_rtype = False




# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Collection of utilities for testing."""

import numpy as np
import numpy.testing
from pint import DimensionalityError

from .units import units


def check_mask(actual, desired):
    """
    Check that two arrays have the same mask.

    This handles the fact that `~numpy.testing.assert_array_equal` ignores masked values
    in either of the arrays. This ensures that the masks are identical.
    """
    actual_mask = getattr(actual, 'mask', np.full(np.asarray(actual).shape, False))
    desired_mask = getattr(desired, 'mask', np.full(np.asarray(desired).shape, False))
    np.testing.assert_array_equal(actual_mask, desired_mask)


def check_and_drop_units(actual, desired):
    """
    Check that the units on the passed in arrays are compatible; return the magnitudes.

    Parameters
    ----------
    actual : `pint.Quantity` or array-like
    desired : `pint.Quantity` or array-like

    Returns
    -------
    actual, desired
        array-like versions of `actual` and `desired` once they have been
        coerced to compatible units.

    Raises
    ------
    AssertionError
        If the units on the passed in objects are not compatible.
    """
    try:
        # If the desired result has units, add dimensionless units if necessary, then
        # ensure that this is compatible to the desired result.
        if hasattr(desired, 'units'):
            if not hasattr(actual, 'units'):
                actual = units.Quantity(actual, 'dimensionless')
            actual = actual.to(desired.units)
        # Otherwise, the desired result has no units. Convert the actual result to
        # dimensionless units if it is a united quantity.
        else:
            if hasattr(actual, 'units'):
                actual = actual.to('dimensionless')
    except DimensionalityError:
        raise AssertionError('Units are not compatible: {} should be {}'.format(
            actual.units, getattr(desired, 'units', 'dimensionless')))

    if hasattr(actual, 'magnitude'):
        actual = actual.magnitude
    if hasattr(desired, 'magnitude'):
        desired = desired.magnitude

    return actual, desired


def assert_array_almost_equal(actual, desired, decimal=7):
    """
    Check that arrays are almost equal, including units.

    Wrapper around :func:`numpy.testing.assert_array_almost_equal`
    """
    actual, desired = check_and_drop_units(actual, desired)
    check_mask(actual, desired)
    numpy.testing.assert_array_almost_equal(actual, desired, decimal)



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""Collection of tools for managing the package."""

# Used to specify functions that should be exported--i.e. added to __all__
# Taken from MetPy's implementation by Unidata and Ryan May
# Inspired by David Beazley and taken from python-ideas:
# https://mail.python.org/pipermail/python-ideas/2014-May/027824.html

__all__ = ('Exporter',)


class Exporter:
    """
    Manages exporting of symbols from the module.

    Grabs a reference to `globals()` for a module and provides a decorator to add
    functions and classes to `__all__` rather than requiring a separately maintained list.
    Also provides a context manager to do this for instances by adding all instances added
    within a block to `__all__`.
    """

    def __init__(self, globls):
        """Initialize the Exporter."""
        self.globls = globls
        self.exports = globls.setdefault('__all__', [])

    def export(self, defn):
        """Declare a function or class as exported."""
        self.exports.append(defn.__name__)
        return defn

    def __enter__(self):
        """Start a block tracking all instances created at global scope."""
        self.start_vars = set(self.globls)

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Exit the instance tracking block."""
        self.exports.extend(set(self.globls) - self.start_vars)
        del self.start_vars



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""
Module to provide unit support.

This makes use of the :mod:`pint` library and sets up the default settings
for good temperature support.

Attributes
----------
units : :class:`pint.UnitRegistry`
    The unit registry used throughout the package. Any use of units in MetPy should
    import this registry and use it to grab units.
"""
import functools
import logging
import re
import warnings

import pint
import pint.unit

log = logging.getLogger(__name__)

UndefinedUnitError = pint.UndefinedUnitError
DimensionalityError = pint.DimensionalityError

# Create registry, with preprocessors for UDUNITS-style powers (m2 s-2) and percent signs
units = pint.UnitRegistry(
    autoconvert_offset_to_baseunit=True,
    preprocessors=[
        functools.partial(
            re.sub,
            r'(?<=[A-Za-z])(?![A-Za-z])(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])',
            '**'
        ),
        lambda string: string.replace('%', 'percent')
    ]
)

# Capture v0.10 NEP 18 warning on first creation
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    units.Quantity([])

# For pint 0.6, this is the best way to define a dimensionless unit. See pint #185
units.define(pint.unit.UnitDefinition('percent', '%', (),
             pint.converters.ScaleConverter(0.01)))

# Silence UnitStrippedWarning
if hasattr(pint, 'UnitStrippedWarning'):
    warnings.simplefilter('ignore', category=pint.UnitStrippedWarning)

# Enable pint's built-in matplotlib support
units.setup_matplotlib()

del pint



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause

"""
===============================
Running the P655 Reduction File
===============================

Experiment p655 was run in 2005 by Mckiernan, Rathbun, and Rowe. It was reduced by Chris
Marone. Here we use the xlook r file parser to run that r file and then get it into the
dictionary of quantity arrays like we use everywhere else and do some simple plotting of the
experiment.

For the curious, this experiment is determining the frictional response of "Ghost Rocks"
from Kodiak Alaska.

First we import a few things so we can get at the test data and we import the xlook parser
object.
"""
##############################
from pylook.cbook import get_test_data
from pylook.io import XlookParser

##############################
# We use pooch to get test data when you run this notebook for the first time, so you won't
# have to use that `get_test_data` function - it's a helper we use to make running pylook
# examples easy! We run it on the r file and the data file so we're sure that both are
# downloaded to your system. In your world, you'll just need to set the `r_file_path` variable
# to the path to your r file. We recommend using pathlib to do this so your code is portable
# across operating systems!

r_file_path = get_test_data('p655_r')

# Getting the l file as well, just so we're sure it's downloaded!
_ = get_test_data('p655intact100l')

##############################
# We need to create an instance of the parser - this is an object that stores all of the
# commands and parsing instructions for interpreting r files and running them.

look = XlookParser()

##############################
# As a hat tip to xlook, we call the `doit` method on our parser to run the r file.
look.doit(r_file_path)

##############################
# Just as xlook did, unknown commands are ignored. In this case we see warnings that the
# strain command is unknown (i.e. we haven't implemented it yet) and there are some follow on
# consequences from that calculation not happening, but we keep running and get valid output!

# The data are currently in a list of arrays in the object, but we want to get the same data
# structure we work with when dealing with data in pure Python - a dictionary of quantity
# arrays! That can be tricky because we need to assign units which are sometimes misspelled
# or just odd. The `get_data_dict` method will do its best, but ultimately fail with unknown
# units. With the `ignore_unknown_units` argument set to `True` it will warn and assign
# dimensionless to anything it doesn't understand. You can also manually specify units for
# all columns, but it is generally easier to fix it up later in practice.

d = look.get_data_dict(ignore_unknown_units=True)

##############################
# Import our unit registry and fix up the bad units to microns as they should have been.
from pylook.units import units

# Fix up that bad unit name
d['ec_disp'] = d['ec_disp'] * units('micron')

##############################
# We'll use Bokeh to take a quick look at the data. Matplotlib is the best choice for your
# publication plots, but the speed and interactivity of Bokeh in the notebook is hard to beat.
# We'll be adding helpers to pylook to make this process easier in the future as well.

# We need to do some imports from bokeh and turn on the notebook backend.

from bokeh.io import output_notebook
from bokeh.layouts import gridplot
from bokeh.plotting import figure, show

output_notebook()

##############################
# This is a handy function that will be integrated into pylook in a more advanced way soon,
# but demonstrates how to make a flexible plotting function instead of copying and pasting a
# bunch of code over and over again.


def make_runplot(data, x_var='Time', y_vars=None,
                 tools='pan,wheel_zoom,box_zoom,reset,save,box_select,hover'):
    plots = []
    for col_name in list(data):
        if col_name == x_var:
            continue
        if y_vars and (col_name not in y_vars):
            continue

        # First plot is simple, the rest we share the x range with the first
        if plots == []:
            p = figure(title=col_name, tools=tools)
        else:
            p = figure(title=col_name, tools=tools, x_range=plots[0].x_range)

        # Plot the data and set the labels
        p.xaxis.axis_label = str(data[x_var].units)
        p.yaxis.axis_label = str(data[col_name].units)
        p.line(data[x_var].m, data[col_name].m)

        plots.append(p)
    show(gridplot(plots, ncols=1, plot_width=600, plot_height=175))

##############################
# By default make_runplot would plot all of the variables, let's just plot a couple of basic
# ones. Hover over the graph to see the values! That can be turned off by clicking the message
# bubble icon in the plot toolbar. If we don't specify, data are plotted with respect to time.


make_runplot(d, y_vars=['Shear_stress', 'Nor_stress'])


##############################
# We can specify to plot relative to another x variable though - with load point displacement
# probably being the most common.

make_runplot(d, x_var='LP_disp', y_vars=['Shear_stress', 'Nor_stress'])

##############################
# That's it! Running the r file for an experiment created for xlook is really just a few lines
# and then we can pull it into the pylook framework easily to manipualte that data with all of
# Python's power. For new experiments, we recommend reducing the experiment in pure Python
# (see other examples), but being able to read and look at older experiments with no fiddling
# is important to utilize the massive amounts of data already collected and reduced.



# Copyright (c) 2020 Leeman Geophysical LLC.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
"""Collection of generally useful utility code from the cookbook."""

import os

import pooch

from . import __version__

POOCH = pooch.create(
    path=pooch.os_cache('pylook'),
    base_url='https://github.com/leemangeophysicalllc/pylook/raw/{version}/staticdata/',
    version='v' + __version__,
    version_dev='master')

# Check if we have the data available directly from a git checkout, either from the
# TEST_DATA_DIR variable, or looking relative to the path of this module's file. Use this
# to override Pooch's path.
dev_data_path = os.environ.get('TEST_DATA_DIR',
                               os.path.join(os.path.dirname(__file__),
                                            '..', 'staticdata'))

if os.path.exists(dev_data_path):
    POOCH.path = dev_data_path

POOCH.load_registry(os.path.join(os.path.dirname(__file__), 'static-data-manifest.txt'))


def get_test_data(fname, as_file_obj=False, mode='rb'):
    """
    Access a file from the collection of test data.

    Parameters
    ----------
    fname : str
        Name of test data file to get
    as_file_obj : boolean
        If the file path should be returned (False, default) or a file like object should be
        opened and returned (True).
    mode : str
        Mode in which to open the file like object.
    """
    path = POOCH.fetch(fname)
    # If we want a file object, open it, trying to guess whether this should be binary mode
    # or not
    if as_file_obj:
        return open(path, mode)

    return path
